import contextlib
import re

import six
from django.db import models, transaction
from django.db.transaction import TransactionManagementError


def create_copy_of_instance(
    instance, attrs=None, exclude=(), save_new=True, using=None
):
    """
    Clone an instance of `django.db.models.Model`.

    :param instance: The model instance to clone.
    :type instance: django.db.models.Model
    :param exclude: List or set of fields to exclude from unique validation.
    :type exclude: list|set
    :param save_new: Save the model instance after duplication calling .save().
    :type save_new: bool
    :param using: The database alias used to save the created instances.
    :type using: str
    :param attrs: Kwargs of field and value to set on the duplicated instance.
    :type attrs: dict
    :return: The new duplicated instance.
    :rtype: django.db.models.Model

    :example:
    >>> from django.contrib.auth import get_user_model
    >>> from sample.models import Book
    >>> instance = Book.objects.create(name='The Beautiful Life')
    >>> instance.pk
    1
    >>> instance.name
    "The Beautiful Life"
    >>> duplicate = create_copy_of_instance(instance, attrs={'name': 'Duplicate Book 2'})
    >>> duplicate.pk
    2
    >>> duplicate.name
    "Duplicate Book 2"
    """

    if not isinstance(instance, models.Model):
        raise ValueError("Invalid: Expected an instance of django.db.models.Model")

    defaults = {}
    attrs = attrs or {}
    default_db_alias = instance._state.db or instance.__class__._default_manager.db
    using = using or default_db_alias
    fields = instance.__class__._meta.concrete_fields

    if not isinstance(attrs, dict):
        try:
            attrs = dict(attrs)
        except (TypeError, ValueError):
            raise ValueError(
                "Invalid: Expected attrs to be a dict or iterable of key and value tuples."
            )

    for f in fields:
        if all(
            [
                not f.auto_created,
                not f.primary_key,
                f.concrete,
                f.editable,
                f not in instance.__class__._meta.related_objects,
                f not in instance.__class__._meta.many_to_many,
            ]
        ):
            # Prevent duplicates
            if f.name not in attrs:
                defaults[f.attname] = getattr(instance, f.attname, f.get_default())

    defaults.update(attrs)

    new_obj = instance.__class__(**defaults)

    exclude = set(
        [
            f.name
            for f in instance._meta.fields
            if any(
                [
                    all([f.name not in defaults, f.attname not in defaults]),
                    f.has_default(),
                    f.null,
                ]
            )
        ]
        + list(exclude)
    )

    # Bug with django using full_clean on a different db
    if using == default_db_alias:
        # Validate the new instance on the same database
        new_obj.full_clean(exclude=exclude)

    if save_new:
        new_obj.save(using=using)

    return new_obj


def unpack_unique_constraints(opts, only_fields=()):
    """
    Unpack unique constraint fields.

    :param opts: Model options
    :type opts: `django.db.models.options.Options`
    :param only_fields: Fields that should be considered.
    :type only_fields: `collections.Iterable`
    :return: Flat list of fields.
    """
    fields = []
    constraints = getattr(
        opts, "total_unique_constraints", getattr(opts, "constraints", [])
    )
    for constraint in constraints:
        fields.extend([f for f in constraint.fields if f in only_fields])
    return fields


def unpack_unique_together(opts, only_fields=()):
    """
    Unpack unique together fields.

    :param opts: Model options
    :type opts: `django.db.models.options.Options`
    :param only_fields: Fields that should be considered.
    :type only_fields: `collections.Iterable`
    :return: Flat list of fields.
    """
    fields = []
    for field in opts.unique_together:
        fields.extend(list([f for f in field if f in only_fields]))
    return fields


def clean_value(value, suffix):
    """
    Strip out copy suffix from a string value.

    :param value: Current value e.g "Test Copy" or "test-copy" for slug fields.
    :type value: `str`
    :param suffix: The suffix value to be replaced with an empty string.
    :type suffix: `str`
    :return: Stripped string without the suffix.
    """
    # type: (str, str) -> str
    return re.sub(r"([\s-]?){}[\s-][\d]$".format(suffix), "", value, flags=re.I)


@contextlib.contextmanager
def transaction_autocommit(using=None):
    """
    Context manager with autocommit enabled.
    """
    try:
        transaction.set_autocommit(True, using=using)
        yield
    except TransactionManagementError:
        raise


@contextlib.contextmanager
def context_mutable_attribute(obj, key, value):
    """
    Context manager that modifies an obj temporarily.
    """
    attribute_exists = hasattr(obj, key)
    default = getattr(obj, key, None)
    try:
        setattr(obj, key, value)
        yield
    finally:
        if attribute_exists:
            setattr(obj, key, default)
        else:
            delattr(obj, key)  # pragma: no cover


def get_value(value, suffix, transform, max_length, index=None):
    """
    Append a suffix to a string value and apply a pass directly to a
    transformation function.
    """
    if index is None:
        duplicate_suffix = " {}".format(suffix.strip())
    else:
        duplicate_suffix = " {} {}".format(suffix.strip(), index)

    total_length = len(value + duplicate_suffix)

    if max_length is not None and total_length > max_length:
        # Truncate the value to max_length - suffix length.
        value = value[: max_length - len(duplicate_suffix)]

    return transform("{}{}".format(value, duplicate_suffix))


def generate_value(value, suffix, transform, max_length, max_attempts):
    """
    Given a fixed max attempt generate a unique value.
    """
    yield get_value(value, suffix, transform, max_length, 1)

    for i in range(1, max_attempts):
        yield get_value(value, suffix, transform, max_length, i)

    raise StopIteration(
        "CloneError: max unique attempts for {} exceeded ({})".format(
            value, max_attempts
        )
    )


def get_unique_value(
    model,
    fname,
    value="",
    transform=lambda v: v,
    suffix="",
    max_length=None,
    max_attempts=100,
):
    """
    Generate a unique value using current value and query the model
    for existing objects with the new value.
    """
    qs = model._default_manager.all()
    it = generate_value(value, suffix, transform, max_length, max_attempts)

    new = six.next(it)
    kwargs = {fname: new}

    while qs.filter(**kwargs).exists():
        new = six.next(it)
        kwargs[fname] = new

    return new


def get_fields_and_unique_fields_from_cls(
    model,
    force,
    clone_fields,
    clone_excluded_fields,
    clone_o2o_fields,
    clone_excluded_o2o_fields,
):
    """Get a list of all fields and unique fields from a model class.

    Skip the clone_* properties if force is ``True``.
    """
    fields = []

    for f in model._meta.concrete_fields:
        valid = False
        if not getattr(f, "primary_key", False):
            if clone_fields and not force and not getattr(f, "one_to_one", False):
                valid = f.name in clone_fields
            elif (
                clone_excluded_fields
                and not force
                and not getattr(f, "one_to_one", False)
            ):
                valid = f.name not in clone_excluded_fields
            elif clone_o2o_fields and not force and getattr(f, "one_to_one", False):
                valid = f.name in clone_o2o_fields
            elif (
                clone_excluded_o2o_fields
                and not force
                and getattr(f, "one_to_one", False)
            ):
                valid = f.name not in clone_excluded_o2o_fields  # pragma: no cover
            else:
                valid = True

        if valid:
            fields.append(f)

    unique_field_names = unpack_unique_together(
        opts=model._meta,
        only_fields=[f.attname for f in fields],
    )

    unique_constraint_field_names = unpack_unique_constraints(
        opts=model._meta,
        only_fields=[f.attname for f in fields],
    )

    unique_fields = [
        f.name
        for f in fields
        if not f.auto_created
        and (
            f.unique
            or f.name in unique_field_names
            or f.name in unique_constraint_field_names
        )
    ]

    return fields, unique_fields


def get_unique_default(
    model,
    fname,
    value,
    transform=lambda v: v,
    suffix="",
    max_length=None,
    max_attempts=100,
):
    """Get a unique value using the value and adding a suffix if needed."""

    qs = model._default_manager.all()

    if not qs.filter(**{fname: value}).exists():
        return value

    it = generate_value(
        value,
        suffix,
        transform,
        max_length,
        max_attempts,
    )

    new = six.next(it)
    kwargs = {fname: new}

    while qs.filter(**kwargs).exists():
        new = six.next(it)
        kwargs[fname] = new

    return new



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import DEFAULT_DB_ALIAS, IntegrityError
from django.test import TestCase
from django.utils.text import slugify

from model_clone import create_copy_of_instance
from sample.models import Book, Library

User = get_user_model()


class CreateCopyOfInstanceTestCase(TestCase):
    REPLICA_DB_ALIAS = "replica"
    databases = {
        "default",
        "replica",
    }

    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create(username="user 1")
        cls.user2 = User.objects.create(username="user 2")

    def test_cloning_model_with_custom_id(self):
        instance = Library.objects.create(name="First library", user=self.user1)
        clone = create_copy_of_instance(instance, attrs={"user": self.user2})

        self.assertNotEqual(instance.pk, clone.pk)
        self.assertEqual(clone.user, self.user2)

    def test_cloning_model_with_a_different_db_alias_is_valid(self):
        new_user = User(username="new user 1")
        new_user.save(using=self.REPLICA_DB_ALIAS)
        instance = Library(name="First library", user=self.user1)
        instance.save(using=DEFAULT_DB_ALIAS)
        clone = create_copy_of_instance(
            instance, attrs={"user": new_user}, using=self.REPLICA_DB_ALIAS
        )

        self.assertNotEqual(instance.pk, clone.pk)
        self.assertEqual(clone.user, new_user)
        self.assertNotEqual(instance._state.db, clone._state.db)

    def test_cloning_unique_fk_field_without_a_fallback_value_is_invalid(self):
        name = "New Library"
        instance = Library.objects.create(name=name, user=self.user1)

        with self.assertRaises(ValidationError):
            create_copy_of_instance(instance)

    def test_cloning_excluded_field_without_a_fallback_value_is_invalid(self):
        name = "New Library"
        instance = Book.objects.create(
            name=name, created_by=self.user1, slug=slugify(name)
        )

        with self.assertRaises(IntegrityError):
            create_copy_of_instance(
                instance, exclude={"slug"}, attrs={"created_by": self.user2}
            )

    def test_raises_error_when_create_copy_of_instance_uses_an_invalid_attrs_value(
        self,
    ):
        instance = Library.objects.create(name="First library", user=self.user1)

        with self.assertRaises(ValueError):
            create_copy_of_instance(instance, attrs="user")

    def test_cloning_an_invalid_object_is_invalid(self):
        class InvalidObj:
            def __init__(self):
                pass

        instance = InvalidObj()

        with self.assertRaises(ValueError):
            create_copy_of_instance(instance, attrs={"created_by": self.user2})



import datetime
import time

from django.contrib.auth import get_user_model
from django.core.checks import Error
from django.core.exceptions import ValidationError
from django.db.transaction import TransactionManagementError
from django.db.utils import DEFAULT_DB_ALIAS, IntegrityError
from django.test import TestCase, TransactionTestCase
from django.utils.text import slugify
from django.utils.timezone import make_naive
from mock import PropertyMock, patch

from model_clone.apps import ModelCloneConfig
from sample.models import (
    Author,
    BackCover,
    Book,
    BookSaleTag,
    BookTag,
    Cover,
    Edition,
    Ending,
    Furniture,
    House,
    Library,
    Page,
    Product,
    Room,
    SaleTag,
    Sentence,
    Tag,
)

User = get_user_model()


class CloneMixinTestCase(TestCase):
    REPLICA_DB_ALIAS = "replica"
    databases = {
        "default",
        "replica",
    }

    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create(username="user 1")
        cls.user2 = User.objects.create(username="user 2")

    def test_cloning_a_transient_instance_with_pk_is_invalid(self):
        instance = Library()

        with self.assertRaises(IntegrityError):
            instance.make_clone()

    def test_cloning_a_transient_instance_is_invalid(self):
        instance = Book()

        with self.assertRaises(ValidationError):
            instance.make_clone()

    def test_cloning_model_with_custom_id(self):
        instance = Library.objects.create(name="First library", user=self.user1)
        clone = instance.make_clone({"user": self.user2})
        clone.refresh_from_db()
        self.assertNotEqual(instance.pk, clone.pk)

    @patch("sample.models.Library._clone_excluded_fields", new_callable=PropertyMock)
    def test_cloning_model_with_excluded_fields(self, _clone_excluded_fields_mock):
        _clone_excluded_fields_mock.return_value = ["name"]
        instance = Library.objects.create(name="First library", user=self.user1)
        clone = instance.make_clone({"name": "New Library"})
        clone.refresh_from_db()
        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.name, clone.name)

    def test_cloning_explict_fields(self):
        name = "New Library"
        instance = Library.objects.create(name=name, user=self.user1)
        clone = instance.make_clone({"user": self.user2, "name": "New name"})
        clone.refresh_from_db()

        self.assertEqual(instance.name, name)
        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.name, clone.name)

    @patch("sample.models.Book._clone_fields", new_callable=PropertyMock)
    def test_cloning_with_default_value(self, _clone_fields_mock):
        instance = Book.objects.create(
            name="Not Published Book", created_by=self.user1, published_at=None
        )
        _clone_fields_mock.return_value = ["name", "created_by", "slug", "published_at"]
        clone = instance.make_clone()
        clone.refresh_from_db()
        self.assertNotEqual(instance.pk, clone.pk)
        self.assertEqual(instance.name, clone.name)
        self.assertNotEqual(instance.slug, clone.slug)
        self.assertIsNone(clone.published_at)

        _clone_fields_mock.return_value.remove("published_at")
        clone = instance.make_clone()
        clone.refresh_from_db()
        self.assertNotEqual(instance.pk, clone.pk)
        self.assertEqual(instance.name, clone.name)
        self.assertNotEqual(instance.slug, clone.slug)
        self.assertIsNotNone(clone.published_at)

    def test_cloning_unique_fk_field(self):
        name = "New Library"
        instance = Library.objects.create(name=name, user=self.user1)
        clone = instance.make_clone({"user": self.user2})
        clone.refresh_from_db()

        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.user, clone.user)

    def test_cloning_unique_o2o_field_without_a_fallback_value_is_valid(self):
        name = "New Library"
        instance = Library.objects.create(name=name, user=self.user1)
        clone = instance.make_clone()
        clone.refresh_from_db()

        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.user.pk, clone.user.pk)

    def test_cloning_related_unique_o2o_field_without_a_fallback_value_is_valid(self):
        with patch(
            "sample.models.Cover._clone_o2o_fields", PropertyMock(return_value=["book"])
        ):
            book = Book.objects.create(
                name="New Book 1", created_by=self.user1, slug=slugify("New Book 1")
            )
            cover = Cover.objects.create(content="New Cover", book=book)
            clone = cover.make_clone()
            clone.refresh_from_db()

            self.assertNotEqual(cover.pk, clone.pk)
            self.assertNotEqual(cover.book.pk, clone.book.pk)
            self.assertEqual(cover.content, clone.content)

        with patch(
            "sample.models.Book._clone_o2o_fields",
            PropertyMock(return_value=["cover", "backcover"]),
        ):
            book = Book.objects.create(
                name="New Book 2", created_by=self.user1, slug=slugify("New Book 2")
            )
            cover = Cover.objects.create(content="New Cover", book=book)
            clone = book.make_clone()
            clone.refresh_from_db()

            self.assertNotEqual(book.pk, clone.pk)
            self.assertNotEqual(cover.pk, clone.cover.pk)
            self.assertEqual(cover.content, clone.cover.content)

            book = Book.objects.create(
                name="New Book 3", created_by=self.user1, slug=slugify("New Book 3")
            )
            backcover = BackCover.objects.create(content="New Back Cover", book=book)
            clone = book.make_clone()
            clone.refresh_from_db()

            self.assertNotEqual(book.pk, clone.pk)
            self.assertIsNotNone(clone.backcover.pk)
            self.assertNotEqual(backcover.pk, clone.backcover.pk)
            self.assertEqual(backcover.content, clone.backcover.content)

    def test_cloning_model_with_a_different_db_alias_is_valid(self):
        name = "New Library"
        instance = Library(name=name, user=self.user1)
        instance.save(using=DEFAULT_DB_ALIAS)
        new_user = User(username="new user 2")
        new_user.save(using=self.REPLICA_DB_ALIAS)
        clone = instance.make_clone(
            attrs={"user": new_user, "name": "New name"},
            using=self.REPLICA_DB_ALIAS,
        )
        clone.refresh_from_db()

        self.assertEqual(instance.name, name)
        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.name, clone.name)

    def test_cloning_with_field_overridden(self):
        name = "New Library"
        instance = Library.objects.create(name=name, user=self.user1)
        new_name = "My New Library"
        clone = instance.make_clone(attrs={"name": new_name, "user": self.user2})
        clone.refresh_from_db()

        self.assertEqual(instance.name, name)

        self.assertNotEqual(instance.pk, clone.pk)
        self.assertNotEqual(instance.name, clone.name)

        self.assertEqual(clone.name, new_name)

    def test_cloning_using_auto_now_field_is_updated(self):
        name = "New Book"
        instance = Book.objects.create(
            name=name, created_by=self.user1, slug=slugify(name)
        )
        time.sleep(1)
        new_name = "My New Book"
        clone = instance.make_clone(attrs={"name": new_name})
        clone.refresh_from_db()

        self.assertEqual(instance.name, name)
        self.assertEqual(clone.created_by, instance.created_by)

        instance_created_at_total_seconds = (
            make_naive(instance.created_at) - datetime.datetime(1970, 1, 1)
        ).total_seconds()

        clone_created_at_total_seconds = (
            make_naive(clone.created_at) - datetime.datetime(1970, 1, 1)
        ).total_seconds()

        self.assertNotEqual(
            instance_created_at_total_seconds,
            clone_created_at_total_seconds,
        )

    def test_cloning_without_explicit_clone_m2m_fields(self):
        author_1 = Author.objects.create(
            first_name="Ruby", last_name="Jack", age=26, sex="F", created_by=self.user1
        )

        author_2 = Author.objects.create(
            first_name="Ibinabo",
            last_name="Jack",
            age=19,
            sex="F",
            created_by=self.user1,
        )

        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))

        book.authors.set([author_1, author_2])

        book_clone = book.make_clone()
        book_clone.refresh_from_db()

        self.assertEqual(book.name, name)
        self.assertEqual(book.created_by, book_clone.created_by)
        self.assertNotEqual(
            list(book.authors.values_list("first_name", "last_name")),
            list(book_clone.authors.values_list("first_name", "last_name")),
        )

    def test_cloning_with_unique_constraint_is_valid(self):
        sale_tag = SaleTag.objects.create(name="test-sale-tag")
        clone_sale_tag_1 = sale_tag.make_clone()
        clone_sale_tag_1.refresh_from_db()

        self.assertNotEqual(sale_tag.pk, clone_sale_tag_1.pk)
        self.assertRegexpMatches(
            clone_sale_tag_1.name,
            r"{}\s[\d]".format(SaleTag.UNIQUE_DUPLICATE_SUFFIX),
        )

        clone_sale_tag_2 = clone_sale_tag_1.make_clone()
        clone_sale_tag_2.refresh_from_db()

        self.assertNotEqual(clone_sale_tag_1.pk, clone_sale_tag_2.pk)
        self.assertRegexpMatches(
            clone_sale_tag_2.name,
            r"{}\s[\d]".format(SaleTag.UNIQUE_DUPLICATE_SUFFIX),
        )

    def test_cloning_with_unique_constraint_uses_field_default(self):
        tag = Tag.objects.create(name="test-tag")
        clone_tag = tag.make_clone()
        clone_tag.refresh_from_db()

        self.assertNotEqual(tag.pk, clone_tag.pk)
        self.assertRegexpMatches(
            clone_tag.name,
            r"\s[\d]",
        )

    @patch("sample.models.Book._clone_m2m_fields", new_callable=PropertyMock)
    def test_cloning_with_explicit_clone_m2m_fields(
        self,
        _clone_m2m_fields_mock,
    ):
        author_1 = Author.objects.create(
            first_name="Opubo", last_name="Jack", age=24, sex="M", created_by=self.user1
        )

        author_2 = Author.objects.create(
            first_name="Nimabo",
            last_name="Jack",
            age=16,
            sex="M",
            created_by=self.user1,
        )
        _clone_m2m_fields_mock.return_value = ["authors"]

        book_1 = Book.objects.create(
            name="New Book 1", created_by=self.user1, slug=slugify("New Book 1")
        )
        book_1.authors.set([author_1, author_2])

        book_clone_1 = book_1.make_clone()
        book_clone_1.refresh_from_db()

        self.assertEqual(
            list(book_1.authors.values_list("first_name", "last_name")),
            list(book_clone_1.authors.values_list("first_name", "last_name")),
        )

        tag_1 = Tag.objects.create(name="test-tag-1")
        tag_2 = Tag.objects.create(name="test-tag-2")

        _clone_m2m_fields_mock.return_value = ["tags"]

        book_2 = Book.objects.create(
            name="New Book 2", created_by=self.user1, slug=slugify("New Book 2")
        )
        BookTag.objects.create(book=book_2, tag=tag_1)
        BookTag.objects.create(book=book_2, tag=tag_2)

        book_clone_2 = book_2.make_clone()
        book_clone_2.refresh_from_db()

        self.assertEqual(
            list(book_2.tags.values_list("name")),
            list(book_clone_2.tags.values_list("name")),
        )

        sale_tag_1 = SaleTag.objects.create(name="test-tag-3")
        sale_tag_2 = SaleTag.objects.create(name="test-tag-4")

        _clone_m2m_fields_mock.return_value = ["sale_tags"]

        book_3 = Book.objects.create(
            name="New Book 3", created_by=self.user1, slug=slugify("New Book 3")
        )
        BookSaleTag.objects.create(book=book_3, sale_tag=sale_tag_1)
        BookSaleTag.objects.create(book=book_3, sale_tag=sale_tag_2)

        book_clone_3 = book_3.make_clone()
        book_clone_3.refresh_from_db()

        self.assertEqual(
            list(book_3.sale_tags.values_list("name")),
            list(book_clone_3.sale_tags.values_list("name")),
        )

    @patch("sample.models.Author._clone_excluded_fields", new_callable=PropertyMock)
    def test_cloning_with_clone_excluded_fields(
        self,
        _clone_excluded_fields_mock,
    ):
        author = Author.objects.create(
            first_name="Opubo", last_name="Jack", age=24, sex="M", created_by=self.user1
        )
        _clone_excluded_fields_mock.return_value = ["last_name"]

        author_clone = author.make_clone()
        author_clone.refresh_from_db()

        self.assertNotEqual(author.last_name, author_clone.last_name)

    @patch("sample.models.Author._clone_m2m_fields", new_callable=PropertyMock)
    def test_cloning_with_explicit_related_clone_m2m_fields(
        self,
        _clone_m2m_fields_mock,
    ):
        author = Author.objects.create(
            first_name="Opubo", last_name="Jack", age=24, sex="M", created_by=self.user1
        )

        _clone_m2m_fields_mock.return_value = ["books"]

        book_1 = Book.objects.create(
            name="New Book 1", created_by=self.user1, slug=slugify("New Book 1")
        )
        book_2 = Book.objects.create(
            name="New Book 2", created_by=self.user1, slug=slugify("New Book 2")
        )
        author.books.set([book_1, book_2])

        author_clone = author.make_clone()
        author_clone.refresh_from_db()

        self.assertEqual(
            list(author.books.values_list("name")),
            list(author_clone.books.values_list("name")),
        )
        _clone_m2m_fields_mock.assert_called()

    def test_cloning_unique_fields_is_valid(self):
        first_name = "Ruby"
        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )

        author_clone = author.make_clone()
        author_clone.refresh_from_db()

        self.assertNotEqual(author.pk, author_clone.pk)
        self.assertEqual(
            author_clone.first_name,
            "{} {} {}".format(first_name, Author.UNIQUE_DUPLICATE_SUFFIX, 1),
        )

    @patch(
        "sample.models.Author.USE_UNIQUE_DUPLICATE_SUFFIX", new_callable=PropertyMock
    )
    def test_cloning_unique_field_with_use_unique_duplicate_suffix_set_to_False(
        self,
        use_unique_duplicate_suffix_mock,
    ):
        use_unique_duplicate_suffix_mock.return_value = False
        first_name = "Ruby"

        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )
        with self.assertRaises(IntegrityError):
            author.make_clone()

        use_unique_duplicate_suffix_mock.assert_called()

    @patch("sample.models.Author.UNIQUE_DUPLICATE_SUFFIX", new_callable=PropertyMock)
    def test_cloning_unique_field_with_a_custom_unique_duplicate_suffix(
        self,
        unique_duplicate_suffix_mock,
    ):
        unique_duplicate_suffix_mock.return_value = "new"
        first_name = "Ruby"

        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )

        author_clone = author.make_clone()
        author_clone.refresh_from_db()

        self.assertNotEqual(author.pk, author_clone.pk)
        self.assertEqual(
            author_clone.first_name,
            "{} {} {}".format(first_name, "new", 1),
        )

    def test_cloning_unique_together_fields_with_enum_field(self):
        first_name = "Ruby"
        last_name = "Jack"

        author = Author.objects.create(
            first_name=first_name,
            last_name=last_name,
            age=26,
            sex="F",
            created_by=self.user1,
        )

        author_clone_1 = author.make_clone()
        author_clone_1.refresh_from_db()

        self.assertNotEqual(author.pk, author_clone_1.pk)
        self.assertEqual(author.sex, author_clone_1.sex)
        self.assertEqual(
            author_clone_1.first_name,
            "{} {} {}".format(first_name, Author.UNIQUE_DUPLICATE_SUFFIX, 1),
        )
        self.assertEqual(
            author_clone_1.last_name,
            Author._meta.get_field("last_name").get_default(),
        )

        author_clone_2 = author.make_clone()
        author_clone_2.refresh_from_db()

        self.assertNotEqual(author.pk, author_clone_2.pk)
        self.assertEqual(author.sex, author_clone_2.sex)
        self.assertEqual(
            author_clone_2.first_name,
            "{} {} {}".format(first_name, Author.UNIQUE_DUPLICATE_SUFFIX, 2),
        )
        self.assertEqual(
            author_clone_2.last_name,
            "{} {} {}".format(
                Author._meta.get_field("last_name").get_default(),
                Author.UNIQUE_DUPLICATE_SUFFIX,
                1,
            ),
        )

        author_clone_3 = author.make_clone()
        author_clone_3.refresh_from_db()

        self.assertNotEqual(author.pk, author_clone_3.pk)
        self.assertEqual(author.sex, author_clone_3.sex)
        self.assertEqual(
            author_clone_3.first_name,
            "{} {} {}".format(first_name, Author.UNIQUE_DUPLICATE_SUFFIX, 3),
        )
        self.assertEqual(
            author_clone_3.last_name,
            "{} {} {}".format(
                Author._meta.get_field("last_name").get_default(),
                Author.UNIQUE_DUPLICATE_SUFFIX,
                2,
            ),
        )

    def test_cloning_unique_slug_field(self):
        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))

        book_clone = book.make_clone()
        book_clone.refresh_from_db()

        self.assertEqual(
            book_clone.slug,
            slugify("{} {} {}".format(book.slug, Book.UNIQUE_DUPLICATE_SUFFIX, 1)),
        )

    def test_making_sub_clones_of_a_unique_slug_field(self):
        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))

        book_clone = book.make_clone()
        book_clone.refresh_from_db()

        self.assertEqual(
            book_clone.slug,
            slugify("{} {} {}".format(book.slug, Book.UNIQUE_DUPLICATE_SUFFIX, 1)),
        )

        for i in range(2, 7):
            book_clone = book_clone.make_clone()
            book_clone.refresh_from_db()

            self.assertEqual(
                book_clone.slug,
                slugify("{} {} {}".format(book.slug, Book.UNIQUE_DUPLICATE_SUFFIX, i)),
            )

    @patch(
        "sample.models.Book.USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS",
        new_callable=PropertyMock,
    )
    def test_making_sub_clones_of_a_non_unique_slug_field_appends_copy(
        self,
        use_duplicate_suffix_for_non_unique_fields_mock,
    ):
        name = "New Book"
        book = Book.objects.create(
            name=name,
            created_by=self.user1,
            slug=slugify(name),
            custom_slug=slugify(name),
        )

        use_duplicate_suffix_for_non_unique_fields_mock.return_value = True

        book_clone = book.make_clone()
        book_clone.refresh_from_db()

        self.assertEqual(
            book_clone.custom_slug,
            slugify("{} {}".format(book.custom_slug, Book.DUPLICATE_SUFFIX)),
        )

    def test_cloning_unique_fields_max_length(self):
        """
        Max unique field length handling.

        Set the initial value for the unique field to max length
        and test to append the [ copy count]
        """
        first_name = (
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, "
            "sed diam nonumy eirmod tempor invidunt ut labore et dolore "
            "magna aliquyam erat, sed diam voluptua. At vero eos et accusam "
            "et justo duo dolores "
        )
        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )

        author_clone = author.make_clone()
        author_clone.refresh_from_db()

        self.assertEqual(
            len(author_clone.first_name),
            Author._meta.get_field("first_name").max_length,
        )
        self.assertNotEqual(author.pk, author_clone.pk)
        self.assertEqual(
            author_clone.first_name,
            "{} {} {}".format(first_name[:193], Author.UNIQUE_DUPLICATE_SUFFIX, 1),
        )

    def test_cloning_instances_in_an_atomic_transaction_with_auto_commit_on_raises_errors(
        self,
    ):
        first_name = (
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, "
            "sed diam nonumy eirmod tempor invidunt ut labore et dolore "
            "magna aliquyam erat, sed diam voluptua. At vero eos et accusam "
            "et justo duo dolores "
        )
        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )

        with self.assertRaises(TransactionManagementError):
            author.bulk_clone(1000, auto_commit=True)

    @patch(
        "sample.models.Book.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS",
        new_callable=PropertyMock,
    )
    def test_bulk_cloning_instances_that_exceed_the_max_unique_count_raises_an_error(
        self,
        max_unique_query_attempts_mock,
    ):
        class InvalidNumber(object):
            """
            Return False for a number that should always be less than the compared value.
            """

            def __init__(self, val):
                self.__val = val

            def __get__(self, instance, owner):
                return self.__val

            def __lt__(self, other):
                return False

            def __gt__(self, other):
                return self.__val > other

            def __ge__(self, other):
                return self.__val > other

        max_unique_query_attempts_mock.return_value = InvalidNumber(100)
        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))

        with self.assertRaises(AssertionError):
            book.bulk_clone(1000)

    def test_cloning_instances_in_an_atomic_transaction_with_auto_commit_off_is_valid(
        self,
    ):
        first_name = (
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, "
            "sed diam nonumy eirmod tempor invidunt ut labore et dolore "
            "magna aliquyam erat, sed diam voluptua. At vero eos et accusam "
            "et justo duo dolores "
        )
        author = Author.objects.create(
            first_name=first_name,
            last_name="Jack",
            age=26,
            sex="F",
            created_by=self.user1,
        )

        clones = author.bulk_clone(1000)

        self.assertEqual(len(clones), 1000)

        for clone in clones:
            self.assertNotEqual(author.pk, clone.pk)
            self.assertRegexpMatches(
                clone.first_name,
                r"{}\s[\d]".format(Author.UNIQUE_DUPLICATE_SUFFIX),
            )

    @patch(
        "sample.models.Book._clone_m2o_or_o2m_fields",
        new_callable=PropertyMock,
    )
    def test_cloning_one_to_many(
        self,
        _clone_m2o_or_o2m_fields_mock,
    ):
        _clone_m2o_or_o2m_fields_mock.return_value = ["page_set"]

        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))

        page_1 = Page.objects.create(content="Page 1 content", book=book)
        page_2 = Page.objects.create(content="Page 2 content", book=book)
        book.page_set.set([page_1, page_2])

        Edition.objects.create(seq=1, book=book)
        self.assertEqual(self.user1.book_set.count(), 1)

        book_clone = book.make_clone()
        book_clone.refresh_from_db()

        self.assertEqual(book.name, name)
        self.assertEqual(book_clone.name, name)
        self.assertEqual(
            list(book.page_set.values_list("content")),
            list(book_clone.page_set.values_list("content")),
        )
        self.assertNotEqual(
            list(book.page_set.values_list("id")),
            list(book_clone.page_set.values_list("id")),
        )
        self.assertNotEqual(book.editions.count(), book_clone.editions.count())
        self.assertEqual(self.user1.book_set.count(), 2)
        _clone_m2o_or_o2m_fields_mock.assert_called()

    @patch(
        "sample.models.Edition._clone_m2o_or_o2m_fields",
        new_callable=PropertyMock,
    )
    def test_cloning_many_to_one(
        self,
        _clone_m2o_or_o2m_fields_mock,
    ):
        _clone_m2o_or_o2m_fields_mock.return_value = ["book"]

        name = "New Book"
        book = Book.objects.create(name=name, created_by=self.user1, slug=slugify(name))
        page = Page.objects.create(content="Page 1 content", book=book)
        edition = Edition.objects.create(seq=1, book=book)

        book.page_set.set([page])
        edition_clone = edition.make_clone()
        edition_clone.refresh_from_db()

        self.assertNotEqual(edition.book.id, edition_clone.book.id)
        self.assertEqual(edition.book.name, edition_clone.book.name)
        _clone_m2o_or_o2m_fields_mock.assert_called()

    @patch(
        "sample.models.Edition._clone_m2o_or_o2m_fields",
        new_callable=PropertyMock,
    )
    def test_cloning_nullable_many_to_one(self, _clone_m2o_or_o2m_fields_mock):
        _clone_m2o_or_o2m_fields_mock.return_value = ["book"]

        edition = Edition.objects.create(seq=1, book=None)
        edition.make_clone()

        self.assertIsNone(edition.book)
        _clone_m2o_or_o2m_fields_mock.assert_called()

    def test_cloning_complex_model_relationships(self):
        house = House.objects.create(name="My House")

        room_1 = Room.objects.create(name="Room 1 in house", house=house)
        room_2 = Room.objects.create(name="Room 2 in house", house=house)

        Furniture.objects.create(name="Chair for room 1", room=room_1)
        Furniture.objects.create(name="Chair for room 2", room=room_2)

        clone_house = house.make_clone()
        clone_house.refresh_from_db()

        self.assertEqual(house.name, clone_house.name)
        self.assertEqual(house.rooms.count(), clone_house.rooms.count())

    @patch(
        "sample.models.Edition.USE_UNIQUE_DUPLICATE_SUFFIX",
        new_callable=PropertyMock,
    )
    @patch(
        "sample.models.Edition.UNIQUE_DUPLICATE_SUFFIX",
        new_callable=PropertyMock,
    )
    def test_unique_duplicate_suffix_check(
        self, unique_duplicate_suffix_mock, use_unique_duplicate_suffix_mock
    ):
        use_unique_duplicate_suffix_mock.return_value = True
        unique_duplicate_suffix_mock.return_value = ""

        errors = Edition.check()
        expected_errors = [
            Error(
                "UNIQUE_DUPLICATE_SUFFIX is required.",
                hint=(
                    "Please provide UNIQUE_DUPLICATE_SUFFIX"
                    + " for {} or set USE_UNIQUE_DUPLICATE_SUFFIX=False".format(
                        Edition.__name__,
                    )
                ),
                obj=Edition,
                id="{}.E001".format(ModelCloneConfig.name),
            )
        ]
        self.assertEqual(errors, expected_errors)

    @patch(
        "sample.models.Edition._clone_fields",
        new_callable=PropertyMock,
    )
    @patch(
        "sample.models.Edition._clone_excluded_fields",
        new_callable=PropertyMock,
    )
    def test_clone_fields_check(self, _clone_excluded_fields_mock, _clone_fields_mock):
        _clone_excluded_fields_mock.return_value = ["test"]
        _clone_fields_mock.return_value = ["test"]

        errors = Edition.check()
        expected_errors = [
            Error(
                "Conflicting configuration.",
                hint=(
                    'Please provide either "_clone_fields"'
                    + ' or "_clone_excluded_fields" for model {}'.format(
                        Edition.__name__,
                    )
                ),
                obj=Edition,
                id="{}.E002".format(ModelCloneConfig.name),
            )
        ]
        self.assertEqual(errors, expected_errors)

    @patch(
        "sample.models.Edition._clone_m2m_fields",
        new_callable=PropertyMock,
    )
    @patch(
        "sample.models.Edition._clone_excluded_m2m_fields",
        new_callable=PropertyMock,
    )
    def test_clone_m2m_fields_check(
        self, _clone_m2m_fields_mock, _clone_excluded_m2m_fields_mock
    ):
        _clone_m2m_fields_mock.return_value = ["test"]
        _clone_excluded_m2m_fields_mock.return_value = ["test"]

        errors = Edition.check()
        expected_errors = [
            Error(
                "Conflicting configuration.",
                hint=(
                    'Please provide either "_clone_m2m_fields"'
                    + ' or "_clone_excluded_m2m_fields" for model {}'.format(
                        Edition.__name__,
                    )
                ),
                obj=Edition,
                id="{}.E002".format(ModelCloneConfig.name),
            )
        ]
        self.assertEqual(errors, expected_errors)

    @patch(
        "sample.models.Edition._clone_m2o_or_o2m_fields",
        new_callable=PropertyMock,
    )
    @patch(
        "sample.models.Edition._clone_excluded_m2o_or_o2m_fields",
        new_callable=PropertyMock,
    )
    def test_clone_m2o_or_o2m_fields_check(
        self, _clone_m2o_or_o2m_fields_mock, _clone_excluded_m2o_or_o2m_fields_mock
    ):
        _clone_m2o_or_o2m_fields_mock.return_value = ["test"]
        _clone_excluded_m2o_or_o2m_fields_mock.return_value = ["test"]

        errors = Edition.check()
        expected_errors = [
            Error(
                "Conflicting configuration.",
                hint=(
                    'Please provide either "_clone_m2o_or_o2m_fields"'
                    + ' or "_clone_excluded_m2o_or_o2m_fields" for model {}'.format(
                        Edition.__name__,
                    )
                ),
                obj=Edition,
                id="{}.E002".format(ModelCloneConfig.name),
            )
        ]
        self.assertEqual(errors, expected_errors)

    @patch(
        "sample.models.Edition._clone_o2o_fields",
        new_callable=PropertyMock,
    )
    @patch(
        "sample.models.Edition._clone_excluded_o2o_fields",
        new_callable=PropertyMock,
    )
    def test_clone_o2o_fields_check(
        self, _clone_o2o_fields_mock, _clone_excluded_o2o_fields_mock
    ):
        _clone_o2o_fields_mock.return_value = ["test"]
        _clone_excluded_o2o_fields_mock.return_value = ["test"]

        errors = Edition.check()
        expected_errors = [
            Error(
                "Conflicting configuration.",
                hint=(
                    'Please provide either "_clone_o2o_fields"'
                    + ' or "_clone_excluded_o2o_fields" for model {}'.format(
                        Edition.__name__,
                    )
                ),
                obj=Edition,
                id="{}.E002".format(ModelCloneConfig.name),
            )
        ]
        self.assertEqual(errors, expected_errors)

    def test_cloning_o2o_fields(self):
        sentence = Sentence.objects.create(value="A really long sentence")
        Ending.objects.create(sentence=sentence)

        self.assertEqual(1, Sentence.objects.count())
        self.assertEqual(1, Ending.objects.count())

        clones = [sentence.make_clone() for _ in range(2)]

        self.assertEqual(2, len(clones))
        self.assertEqual(3, Sentence.objects.count())


class CloneMixinTransactionTestCase(TransactionTestCase):
    def test_cloning_multiple_instances_doesnt_exceed_the_max_length(self):
        user = User.objects.create()
        first_name = (
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, "
            "sed diam nonumy eirmod tempor invidunt ut labore et dolore "
            "magna aliquyam erat, sed diam voluptua. At vero eos et accusam "
            "et justo duo dolores "
        )
        author = Author.objects.create(
            first_name=first_name, last_name="Jack", age=26, sex="F", created_by=user
        )

        clones = author.bulk_clone(1000)

        self.assertEqual(len(clones), 1000)

        for clone in clones:
            self.assertNotEqual(author.pk, clone.pk)
            self.assertRegexpMatches(
                clone.first_name,
                r"{}\s[\d]".format(Author.UNIQUE_DUPLICATE_SUFFIX),
            )

    def test_cloning_multiple_instances_with_autocommit_is_valid(self):
        user = User.objects.create()
        first_name = (
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, "
            "sed diam nonumy eirmod tempor invidunt ut labore et dolore "
            "magna aliquyam erat, sed diam voluptua. At vero eos et accusam "
            "et justo duo dolores 2"
        )
        author = Author.objects.create(
            first_name=first_name, last_name="Jack", age=26, sex="F", created_by=user
        )

        clones = author.bulk_clone(1000, auto_commit=True)

        self.assertEqual(len(clones), 1000)

        for clone in clones:
            self.assertNotEqual(author.pk, clone.pk)
            self.assertRegexpMatches(
                clone.first_name,
                r"{}\s[\d]".format(Author.UNIQUE_DUPLICATE_SUFFIX),
            )

    def test_cloning_model_with_unique_text_field(self):
        product = Product.objects.create(name="Test Product")
        clone = product.make_clone()
        clone.refresh_from_db()
        self.assertEqual(
            clone.name,
            "{0} {1} 1".format(product.name, Product.UNIQUE_DUPLICATE_SUFFIX),
        )



from django.contrib.admin import ModelAdmin
from django.contrib.admin.exceptions import DisallowedModelAdminToField
from django.contrib.admin.options import TO_FIELD_VAR
from django.contrib.admin.utils import unquote
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from model_clone.mixin import CloneMixin


class CloneModelAdminMixin(object):
    """Mixin to handle duplication of models."""

    include_duplicate_action = True
    include_duplicate_object_link = True

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context[
            "include_duplicate_object_link"
        ] = self.include_duplicate_object_link
        if self.include_duplicate_object_link:
            to_field = request.POST.get(TO_FIELD_VAR, request.GET.get(TO_FIELD_VAR))
            if to_field and not self.to_field_allowed(request, to_field):
                raise DisallowedModelAdminToField(
                    "The field %s cannot be referenced." % to_field
                )

            obj = self.get_object(request, unquote(object_id), to_field)

            if obj is not None and request.GET.get("duplicate") == "true":
                clone = obj.make_clone()
                clone.save()
                self.message_user(
                    request,
                    _("Duplication successful, redirected to cloned: {}").format(clone),
                )
                cloned_admin_url = reverse(
                    "admin:{0}_{1}_change".format(
                        clone._meta.app_label, clone._meta.model_name
                    ),
                    args=(clone.pk,),
                )
                return HttpResponseRedirect(cloned_admin_url)

        return super().changeform_view(request, object_id, form_url, extra_context)

    def make_clone(self, request, queryset):
        clone_obj_ids = []

        for obj in queryset:
            clone = obj.make_clone()
            clone.save()
            clone_obj_ids.append(str(clone.pk))

        if clone_obj_ids:
            self.message_user(
                request,
                _("Successfully created: {} new duplicates").format(len(clone_obj_ids)),
            )

    make_clone.short_description = _("Duplicate selected %(verbose_name_plural)s")

    def _get_base_actions(self):
        """Return the list of actions, prior to any request-based filtering."""
        actions = list(super()._get_base_actions())
        # Add the make clone action
        if self.include_duplicate_action and issubclass(self.model, CloneMixin):
            actions.extend(self.get_action(action) for action in ["make_clone"])

        return actions


class CloneModelAdmin(CloneModelAdminMixin, ModelAdmin):
    """Clone model admin view."""



import io
import os

from setuptools import find_packages, setup

install_requires = [
    "django",
    "future>=0.17.1",
    "conditional>=1.3",
    "six",
]

test_requires = [
    "tox",
    "tox-gh-actions",
    "pluggy>=0.7",
    "mock",
    "unittest-xml-reporting",
    "codacy-coverage",
    "django-migration-fixer",
]

deploy_requires = [
    "bump2version",
    "readme_renderer[md]",
    "changes",
    "git-changelog",
    "twine==4.0.1",
]

lint_requires = [
    "flake8",
    "yamllint",
    "isort",
    "black",
]

local_dev_requires = [
    "pip-tools",
    "check-manifest",
]

extras_require = {
    "development": [
        local_dev_requires,
        install_requires,
        test_requires,
        lint_requires,
        deploy_requires,
    ],
    "test": test_requires,
    "lint": lint_requires,
    "deploy": deploy_requires,
}

BASE_DIR = os.path.dirname(__file__)
README_PATH = os.path.join(BASE_DIR, "README.md")
LONG_DESCRIPTION_TYPE = "text/markdown"

if os.path.isfile(README_PATH):
    with io.open(README_PATH, encoding="utf-8") as f:
        LONG_DESCRIPTION = f.read()
else:
    LONG_DESCRIPTION = ""

setup(
    name="django-clone",
    version="3.0.5",
    description="Create a clone of a django model instance.",
    python_requires=">=3.6",
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_TYPE,
    author="Tonye Jack",
    author_email="jtonye@ymail.com",
    maintainer="Tonye Jack",
    maintainer_email="jtonye@ymail.com",
    url="https://github.com/tj-django/django-clone.git",
    license="MIT/Apache-2.0",
    zip_safe=False,
    include_package_data=True,
    keywords=[
        "django",
        "django-clone",
        "django clone",
        "django object clone",
        "clone-django",
        "model cloning",
        "django instance duplication",
        "django duplication",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Topic :: Internet :: WWW/HTTP",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
    ],
    install_requires=install_requires,
    tests_require=["coverage"],
    extras_require=extras_require,
    packages=find_packages(
        include=["model_clone"], exclude=["tests.*", "tests", "demo"]
    ),
)



"""Top-level package for django-clone."""

__author__ = """Tonye Jack"""
__email__ = "jtonye@ymail.com"
__version__ = "3.0.5"

from model_clone.admin import CloneModelAdmin, CloneModelAdminMixin
from model_clone.mixin import CloneMixin
from model_clone.utils import create_copy_of_instance

__all__ = [
    "CloneMixin",
    "CloneModelAdmin",
    "CloneModelAdminMixin",
    "create_copy_of_instance",
]



import itertools
from itertools import repeat
from typing import Dict, List, Optional

from conditional import conditional
from django.core.checks import Error
from django.core.exceptions import ValidationError
from django.db import IntegrityError, connections, models, transaction
from django.db.models import SlugField
from django.utils.text import slugify

from model_clone.apps import ModelCloneConfig
from model_clone.utils import (
    clean_value,
    context_mutable_attribute,
    get_fields_and_unique_fields_from_cls,
    get_unique_default,
    get_unique_value,
    get_value,
    transaction_autocommit,
)


class CloneMixin(object):

    """CloneMixin mixin to duplicate an object using the model's class.

    :param _clone_fields: Restricted List of fields to copy from the instance.
    :type _clone_fields: collections.Iterable
    :param _clone_m2m_fields: Many to many fields (Example: TestModel.tags).
    :type _clone_m2m_fields: collections.Iterable
    :param _clone_m2o_or_o2m_fields: Many to one/One to many fields.
    :type _clone_m2o_or_o2m_fields: collections.Iterable
    :param _clone_o2o_fields: One to One fields.
    :type _clone_o2o_fields: collections.Iterable

    :param _clone_excluded_fields: Excluded model fields.
    :type _clone_excluded_fields: collections.Iterable
    :param _clone_excluded_m2m_fields: Excluded many to many fields.
    :type _clone_excluded_m2m_fields: collections.Iterable
    :param _clone_excluded_m2o_or_o2m_fields: Excluded many to many
        and one to many fields.
    :type _clone_excluded_m2o_or_o2m_fields: collections.Iterable
    :param _clone_excluded_o2o_fields: Excluded one to one fields.
    :type _clone_excluded_o2o_fields: collections.Iterable

    :Example:
    >>> from django.conf import settings
    >>> from django.db import models
    >>>
    >>> # Using explicit fields
    >>>
    >>> class TestModel1(CloneMixin, models.Model):
    >>>     field_1 = models.CharField(max_length=200)
    >>>     tags =  models.ManyToManyField(Tags)
    >>>     audiences = models.ManyToManyField(Audience)
    >>>     user = models.ForiegnKey(
    >>>         settings.AUTH_USER_MODEL,
    >>>         on_delete=models.CASCADE,
    >>>     )
    >>>
    >>>     _clone_m2m_fields = ['tags', 'audiences']
    >>>     _clone_m2o_or_o2m_fields = ['user']
    >>>     ...
    >>>
    >>> # Using implicit all except fields.
    >>>
    >>> class TestModel2(CloneMixin, models.Model):
    >>>     field_1 = models.CharField(max_length=200)
    >>>     tags =  models.ManyToManyField(Tags)
    >>>     audiences = models.ManyToManyField(Audience)
    >>>     user = models.ForiegnKey(
    >>>         settings.AUTH_USER_MODEL,
    >>>         on_delete=models.CASCADE,
    >>>         null=True,
    >>>     )

    >>>     # Clones any other m2m field excluding "audiences".
    >>>     _clone_excluded_m2m_fields = ['audiences']
    >>>     # Clones all other many to one or one to many fields excluding "user".
    >>>     _clone_excluded_m2o_or_o2m_fields = ['user']
    >>>     ...
    """

    # Included fields
    _clone_fields = []  # type: List[str]
    _clone_m2m_fields = []  # type: List[str]
    _clone_m2o_or_o2m_fields = []  # type: List[str]
    _clone_o2o_fields = []  # type: List[str]

    # Excluded fields
    _clone_excluded_fields = []  # type: List[str]
    _clone_excluded_m2m_fields = []  # type: List[str]
    _clone_excluded_m2o_or_o2m_fields = []  # type: List[str]
    _clone_excluded_o2o_fields = []  # type: List[str]

    DUPLICATE_SUFFIX = "copy"  # type: str
    USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS = False  # type: bool
    UNIQUE_DUPLICATE_SUFFIX = "copy"  # type: str
    USE_UNIQUE_DUPLICATE_SUFFIX = True  # type: bool
    MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS = 100  # type: int

    @classmethod
    def check(cls, **kwargs):  # pragma: no cover
        errors = super(CloneMixin, cls).check(**kwargs)

        if cls.USE_UNIQUE_DUPLICATE_SUFFIX and not cls.UNIQUE_DUPLICATE_SUFFIX:
            errors.append(
                Error(
                    "UNIQUE_DUPLICATE_SUFFIX is required.",
                    hint=(
                        "Please provide UNIQUE_DUPLICATE_SUFFIX"
                        f" for {cls.__name__} or set USE_UNIQUE_DUPLICATE_SUFFIX=False"
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E001",
                )
            )

        if cls.USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS and not cls.DUPLICATE_SUFFIX:
            errors.append(
                Error(
                    "UNIQUE_DUPLICATE_SUFFIX is required.",
                    hint=(
                        f"Please provide DUPLICATE_SUFFIX for {cls.__name__} "
                        "or set USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS=False"
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E001",
                )
            )

        if all([cls._clone_fields, cls._clone_excluded_fields]):
            errors.append(
                Error(
                    "Conflicting configuration.",
                    hint=(
                        'Please provide either "_clone_fields"'
                        f' or "_clone_excluded_fields" for model {cls.__name__}'
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E002",
                )
            )

        if all([cls._clone_m2m_fields, cls._clone_excluded_m2m_fields]):
            errors.append(
                Error(
                    "Conflicting configuration.",
                    hint=(
                        'Please provide either "_clone_m2m_fields"'
                        f' or "_clone_excluded_m2m_fields" for model {cls.__name__}'
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E002",
                )
            )

        if all(
            [
                cls._clone_m2o_or_o2m_fields,
                cls._clone_excluded_m2o_or_o2m_fields,
            ]
        ):
            errors.append(
                Error(
                    "Conflicting configuration.",
                    hint=(
                        'Please provide either "_clone_m2o_or_o2m_fields" or '
                        f'"_clone_excluded_m2o_or_o2m_fields" for model {cls.__name__}'
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E002",
                )
            )

        if all([cls._clone_o2o_fields, cls._clone_excluded_o2o_fields]):
            errors.append(
                Error(
                    "Conflicting configuration.",
                    hint=(
                        'Please provide either "_clone_o2o_fields" or '
                        f'"_clone_excluded_o2o_fields" for model {cls.__name__}'
                    ),
                    obj=cls,
                    id=f"{ModelCloneConfig.name}.E002",
                )
            )

        return errors

    @classmethod
    def bulk_clone_multi(cls, objs, attrs=None, batch_size=None):
        # type: (List[models.Model], Optional[List[Dict]], Optional[int]) -> List[models.Model]
        # TODO: Support bulk clones split by the batch_size
        pass

    def pre_save_duplicate(self, instance):  # pylint: disable=R0201
        """Override this method to modify the duplicate instance before it's saved."""
        return instance

    @transaction.atomic
    def make_clone(self, attrs=None, sub_clone=False, using=None):
        """Creates a clone of the django model instance.

        :param attrs: Dictionary of attributes to be replaced on the cloned object.
        :type attrs: dict
        :param sub_clone: Internal boolean used to detect cloning sub objects.
        :type sub_clone: bool
        :rtype: :obj:`django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :return: The model instance that has been cloned.
        """
        using = using or self._state.db or self.__class__._default_manager.db
        attrs = attrs or {}
        if not self.pk:
            raise ValidationError(
                "{}: Instance must be saved before it can be cloned.".format(
                    self.__class__.__name__
                )
            )
        if sub_clone:
            duplicate = self  # pragma: no cover
            duplicate.pk = None  # pragma: no cover
        else:
            duplicate = self._create_copy_of_instance(self, using=using)

        for name, value in attrs.items():
            setattr(duplicate, name, value)

        duplicate = self.pre_save_duplicate(duplicate)
        duplicate.save(using=using)

        duplicate = self.__duplicate_m2o_fields(duplicate, using=using)
        duplicate = self.__duplicate_o2o_fields(duplicate, using=using)
        duplicate = self.__duplicate_o2m_fields(duplicate, using=using)
        duplicate = self.__duplicate_m2m_fields(duplicate, using=using)

        duplicate.save(using=using)

        return duplicate

    def bulk_clone(
        self, count, attrs=None, batch_size=None, using=None, auto_commit=False
    ):
        using = using or self._state.db or self.__class__._default_manager.db
        ops = connections[using].ops
        objs = range(count)
        clones = []
        batch_size = batch_size or max(ops.bulk_batch_size([], list(objs)), 1)

        with conditional(
            auto_commit,
            transaction_autocommit(using=using),
        ):
            # If count exceeds the MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS
            with conditional(
                self.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS < count,
                context_mutable_attribute(
                    self,
                    "MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS",
                    count,
                ),
            ):
                if not self.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS >= count:
                    raise AssertionError(
                        "An Unknown error has occurred: Expected ({}) >= ({})".format(
                            self.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS, count
                        ),
                    )
                clones = list(repeat(self.make_clone(attrs=attrs), batch_size))

        return clones

    def parallel_clone(self, count, attrs=None, batch_size=None, auto_commit=False):
        # if this takes n time for t records
        # t^n i.e 100 ** 2ms = 10000ms (10s) to clone 100 objects.
        # I'll like to reduce this down to max time to clone count/batch_size i.e
        # If it take 10000ms to clone 100 objects with a db of batch_size 100
        # If it takes 20ms to clone 10 objects I'll like to keep this down to 20ms
        # i.e max_num_of_threads 10 for 100 objects.
        # This should run in parallel
        # Testing jit and cpython if they offer better API's.
        pass

    @staticmethod
    def _create_copy_of_instance(instance, using=None, force=False, sub_clone=False):
        """Create a copy of a model instance.

        :param instance: The instance to be duplicated.
        :type instance: `django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :param force: Flag to skip using the current model clone declared attributes.
        :type force: bool
        :param sub_clone: Flag to skip cloning one to one fields for sub clones.
        :type sub_clone: bool
        :return: A new transient instance.
        :rtype: `django.db.models.Model`
        """
        cls = instance.__class__
        clone_fields = getattr(cls, "_clone_fields", CloneMixin._clone_fields)
        clone_excluded_fields = getattr(
            cls, "_clone_excluded_fields", CloneMixin._clone_excluded_fields
        )
        clone_o2o_fields = getattr(
            cls, "_clone_o2o_fields", CloneMixin._clone_o2o_fields
        )
        clone_excluded_o2o_fields = getattr(
            cls, "_clone_excluded_o2o_fields", CloneMixin._clone_excluded_o2o_fields
        )
        unique_duplicate_suffix = getattr(
            cls, "UNIQUE_DUPLICATE_SUFFIX", CloneMixin.UNIQUE_DUPLICATE_SUFFIX
        )
        use_unique_duplicate_suffix = getattr(
            cls,
            "USE_UNIQUE_DUPLICATE_SUFFIX",
            CloneMixin.USE_UNIQUE_DUPLICATE_SUFFIX,
        )
        max_unique_duplicate_query_attempts = getattr(
            cls,
            "MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS",
            CloneMixin.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS,
        )
        duplicate_suffix = getattr(
            cls,
            "DUPLICATE_SUFFIX",
            CloneMixin.DUPLICATE_SUFFIX,
        )
        use_duplicate_suffix_for_non_unique_fields = getattr(
            cls,
            "USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS",
            CloneMixin.USE_DUPLICATE_SUFFIX_FOR_NON_UNIQUE_FIELDS,
        )

        fields, unique_fields = get_fields_and_unique_fields_from_cls(
            model=cls,
            force=force,
            clone_fields=clone_fields,
            clone_excluded_fields=clone_excluded_fields,
            clone_o2o_fields=clone_o2o_fields,
            clone_excluded_o2o_fields=clone_excluded_o2o_fields,
        )

        new_instance = cls()

        for f in fields:
            if not f.editable:
                f.pre_save(new_instance, add=True)
                continue

            value = getattr(instance, f.attname)

            if all(
                [
                    not f.auto_created,
                    f.concrete,
                    f.editable,
                    f.name in unique_fields,
                ]
            ):
                # Do not try to get unique value for enum type field
                if all(
                    [
                        isinstance(f, (models.CharField, models.TextField)),
                        not f.choices,
                        isinstance(value, str),
                    ]
                ):
                    value = clean_value(value, unique_duplicate_suffix)

                    if f.has_default():
                        value = f.get_default()

                        if not callable(f.default) and isinstance(value, str):
                            value = get_unique_default(
                                model=cls,
                                fname=f.attname,
                                value=value,
                                transform=(
                                    slugify if isinstance(f, SlugField) else str
                                ),
                                suffix=unique_duplicate_suffix,
                                max_length=f.max_length,
                                max_attempts=max_unique_duplicate_query_attempts,
                            )

                    elif use_unique_duplicate_suffix and isinstance(value, str):
                        value = get_unique_value(
                            model=cls,
                            fname=f.attname,
                            value=value,
                            transform=(slugify if isinstance(f, SlugField) else str),
                            suffix=unique_duplicate_suffix,
                            max_length=f.max_length,
                            max_attempts=max_unique_duplicate_query_attempts,
                        )

                elif isinstance(f, models.OneToOneField) and not sub_clone:
                    sub_instance = getattr(instance, f.name, None) or f.get_default()

                    if sub_instance is not None:
                        sub_instance = CloneMixin._create_copy_of_instance(
                            sub_instance,
                            force=True,
                            sub_clone=True,
                        )
                        sub_instance.save(using=using)
                        value = sub_instance.pk
            elif all(
                [
                    use_duplicate_suffix_for_non_unique_fields,
                    f.concrete,
                    f.editable,
                    f.name not in unique_fields,
                ]
            ):
                if (
                    isinstance(f, (models.CharField, models.TextField))
                    and not f.choices
                ):
                    value = get_value(
                        value=value,
                        transform=(slugify if isinstance(f, SlugField) else str),
                        suffix=duplicate_suffix,
                        max_length=f.max_length,
                    )

            setattr(new_instance, f.attname, value)

        return new_instance

    def __duplicate_o2o_fields(self, duplicate, using=None):
        """Duplicate one to one fields.
        :param duplicate: The transient instance that should be duplicated.
        :type duplicate: `django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :return: The duplicate instance with all the one to one fields duplicated.
        """
        for f in self._meta.related_objects:
            if f.one_to_one:
                if any(
                    [
                        f.name in self._clone_o2o_fields
                        and f not in self._meta.concrete_fields,
                        self._clone_excluded_o2o_fields
                        and f.name not in self._clone_excluded_o2o_fields
                        and f not in self._meta.concrete_fields,
                    ]
                ):
                    rel_object = getattr(self, f.name, None)
                    if rel_object:
                        new_rel_object = CloneMixin._create_copy_of_instance(
                            rel_object,
                            force=True,
                            sub_clone=True,
                        )
                        setattr(new_rel_object, f.remote_field.name, duplicate)
                        new_rel_object.save(using=using)

        return duplicate

    def __duplicate_o2m_fields(self, duplicate, using=None):
        """Duplicate one to many fields.

        :param duplicate: The transient instance that should be duplicated.
        :type duplicate: `django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :return: The duplicate instance with all the transient one to many duplicated instances.
        """

        for f in itertools.chain(
            self._meta.related_objects, self._meta.concrete_fields
        ):
            if f.one_to_many:
                if any(
                    [
                        f.get_accessor_name() in self._clone_m2o_or_o2m_fields,
                        self._clone_excluded_m2o_or_o2m_fields
                        and f.get_accessor_name()
                        not in self._clone_excluded_m2o_or_o2m_fields,
                    ]
                ):
                    for item in getattr(self, f.get_accessor_name()).all():
                        if hasattr(item, "make_clone"):
                            try:
                                item.make_clone(
                                    attrs={f.remote_field.name: duplicate},
                                    using=using,
                                )
                            except IntegrityError:
                                item.make_clone(
                                    attrs={f.remote_field.name: duplicate},
                                    sub_clone=True,
                                    using=using,
                                )
                        else:
                            new_item = CloneMixin._create_copy_of_instance(
                                item,
                                force=True,
                                sub_clone=True,
                                using=using,
                            )
                            setattr(new_item, f.remote_field.name, duplicate)

                            new_item.save(using=using)

        return duplicate

    def __duplicate_m2o_fields(self, duplicate, using=None):
        """Duplicate many to one fields.

        :param duplicate: The transient instance that should be duplicated.
        :type duplicate: `django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :return: The duplicate instance with all the many to one fields duplicated.
        """
        for f in self._meta.concrete_fields:
            if f.many_to_one:
                if any(
                    [
                        f.name in self._clone_m2o_or_o2m_fields,
                        self._clone_excluded_m2o_or_o2m_fields
                        and f.name not in self._clone_excluded_m2o_or_o2m_fields,
                    ]
                ):
                    item = getattr(self, f.name)
                    if hasattr(item, "make_clone"):
                        try:
                            item_clone = item.make_clone(using=using)
                        except IntegrityError:
                            item_clone = item.make_clone(sub_clone=True)
                    elif item is None:
                        item_clone = None
                    else:
                        item.pk = None  # pragma: no cover
                        item_clone = item.save(using=using)  # pragma: no cover

                    setattr(duplicate, f.name, item_clone)

        return duplicate

    def __duplicate_m2m_fields(self, duplicate, using=None):
        """Duplicate many to many fields.

        :param duplicate: The transient instance that should be duplicated.
        :type duplicate: `django.db.models.Model`
        :param using: The database alias used to save the created instances.
        :type using: str
        :return: The duplicate instance with all the many to many fields duplicated.
        """
        fields = set()

        for f in self._meta.many_to_many:
            if any(
                [
                    f.name in self._clone_m2m_fields,
                    self._clone_excluded_m2m_fields
                    and f.name not in self._clone_excluded_m2m_fields,
                ]
            ):
                fields.add(f)

        for f in self._meta.related_objects:
            if f.many_to_many:
                if any(
                    [
                        f.get_accessor_name() in self._clone_m2m_fields,
                        self._clone_excluded_m2m_fields
                        and f.get_accessor_name()
                        not in self._clone_excluded_m2m_fields,
                    ]
                ):
                    fields.add(f)

        # Clone many to many fields
        for field in fields:
            if hasattr(field, "field"):
                # ManyToManyRel
                field_name = field.field.m2m_reverse_field_name()
                through = field.through
                source = getattr(self, field.get_accessor_name())
                destination = getattr(duplicate, field.get_accessor_name())
            else:
                through = field.remote_field.through
                field_name = field.m2m_field_name()
                source = getattr(self, field.attname)
                destination = getattr(duplicate, field.attname)
            if all(
                [
                    through,
                    not through._meta.auto_created,
                ]
            ):
                objs = through.objects.filter(**{field_name: self.pk})
                for item in objs:
                    if hasattr(through, "make_clone"):
                        try:
                            item.make_clone(
                                attrs={field_name: duplicate},
                                using=using,
                            )
                        except IntegrityError:
                            item.make_clone(
                                attrs={field_name: duplicate},
                                sub_clone=True,
                                using=using,
                            )
                    else:
                        item.pk = None
                        setattr(item, field_name, duplicate)
                        item.save(using=using)
            else:
                destination.set(source.all())

        return duplicate
